// ********************************
// * Human-like Bot Aiming Plugin *
// ********************************
// version 1.5
// by Pierre-Marie Baty <pm@bots-united.com>


#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

#define AIM_GENUINE 0
#define AIM_RACC_OLD 1
#define AIM_RACC 2


enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable_Post;
cvar_t bot_turn_method = {"bot_turn_method", "2", FCVAR_EXTDLL};
cvar_t bot_turn_skill = {"bot_turn_skill", "4", FCVAR_EXTDLL};
cvar_t bot_aim_fix = {"bot_aim_fix", "1", FCVAR_EXTDLL};
int turn_method;
bool aim_fix;
int turn_skill;
Vector bot_v_angle[32];

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
void StartFrame_Post (void);
void TraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr);
void FakeclientPointGun (edict_t *pEdict, bool save_angles);
bool AimsAtSomething (edict_t *pEdict);
float UTIL_AngleOfVectors (Vector v1, Vector v2);
float UTIL_WrapAngle (float angle_to_wrap);
Vector UTIL_WrapAngles (Vector angles_to_wrap);


// START of Metamod stuff

enginefuncs_t meta_engfuncs;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   NULL, // pfnGetEntityAPI2()
   GetEntityAPI2_Post, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   GetEngineFunctions, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "BotAim", // plugin name
   "1.5", // plugin version
   "22/05/2003", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "BOTAIM", // plugin logtag
   PT_ANYTIME, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&bot_turn_method);
   CVAR_REGISTER (&bot_aim_fix);
   CVAR_REGISTER (&bot_turn_skill);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2_Post (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   gFunctionTable_Post.pfnStartFrame = StartFrame_Post;

   memcpy (pFunctionTable, &gFunctionTable_Post, sizeof (DLL_FUNCTIONS));
   return (TRUE);
}


C_DLLEXPORT int GetEngineFunctions (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   meta_engfuncs.pfnTraceLine = TraceLine;

   memcpy (pengfuncsFromEngine, &meta_engfuncs, sizeof (enginefuncs_t));
   return TRUE;
}

// END of Metamod stuff




#ifdef _WIN32
BOOL WINAPI DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   return TRUE; // required Win32 DLL entry point
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


void StartFrame_Post (void)
{
   int player_index;

   // get the turn method and do a fool-proof check
   turn_method = CVAR_GET_FLOAT ("bot_turn_method");
   if (turn_method < 0) { CVAR_SET_FLOAT ("bot_turn_method", 0); turn_method = 0; }
   else if (turn_method > 2) { CVAR_SET_FLOAT ("bot_turn_method", 2); turn_method = 2; }

   // get the turn skill and do a fool-proof check
   turn_skill = CVAR_GET_FLOAT ("bot_turn_skill");
   if (turn_skill < 1) { CVAR_SET_FLOAT ("bot_turn_skill", 1); turn_skill = 1; }

   // get whether we need to fix "aimbug fix" bug or not
   aim_fix = (CVAR_GET_FLOAT ("bot_aim_fix") > 0);

   // cycle through all players
   for (player_index = 1; player_index <= gpGlobals->maxClients; player_index++)
   {
      edict_t *pPlayer = INDEXENT (player_index);

      if (FNullEnt (pPlayer) || pPlayer->free)
         continue; // skip invalid players

      // if this player is a bot AND user specified a custom aim method, make it aim and turn
      if ((pPlayer->v.flags & FL_FAKECLIENT) && (turn_method != AIM_GENUINE))
         FakeclientPointGun (pPlayer, TRUE); // update and save this bot's view angles

      // systematically wrap all player angles to avoid engine freezes
      pPlayer->v.angles = UTIL_WrapAngles (pPlayer->v.angles);
      pPlayer->v.v_angle = UTIL_WrapAngles (pPlayer->v.v_angle);
      pPlayer->v.idealpitch = UTIL_WrapAngle (pPlayer->v.idealpitch);
      pPlayer->v.ideal_yaw = UTIL_WrapAngle (pPlayer->v.ideal_yaw);
   }

   RETURN_META (MRES_IGNORED);
}


void TraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr)
{
   // check to see if a bot is sending this traceline AND it is firing AND we are taking over the bot aiming code
   if (!FNullEnt (pentToSkip) && (pentToSkip->v.flags & FL_FAKECLIENT)
       && (pentToSkip->v.button & (IN_ATTACK | IN_ATTACK2)) && (turn_method != AIM_GENUINE))
   {
      FakeclientPointGun (pentToSkip, FALSE); // update the bot's view angles in advance, but don't save them
      MAKE_VECTORS (pentToSkip->v.v_angle); // build base vectors in bot's actual view direction

      // trace the line considering the ACTUAL view angles of the bot, not the ones it claims to have
      TRACE_LINE (pentToSkip->v.origin + pentToSkip->v.view_ofs,
                  pentToSkip->v.origin + pentToSkip->v.view_ofs + gpGlobals->v_forward * 9999,
                  fNoMonsters, pentToSkip, ptr);

      RETURN_META (MRES_SUPERCEDE); // and don't let the engine be fooled by the bot's lies, goddamnit! :)
   }

   RETURN_META (MRES_IGNORED);
}


void FakeclientPointGun (edict_t *pEdict, bool save_angles)
{
   // this function is called every frame for every bot. Its purpose is to make the bot
   // move its crosshair to the direction where it wants to look. There is some kind of
   // filtering for the view, to make it human-like.

   float speed; // speed : 0.1 - 1
   Vector v_deviation;

   v_deviation = UTIL_WrapAngles (Vector (pEdict->v.idealpitch, pEdict->v.ideal_yaw, 0) - bot_v_angle[ENTINDEX (pEdict) - 1]);

   // do we want old RACC style aiming ?
   if (turn_method == AIM_RACC_OLD)
   {
      // limit the turn skill
      speed = (turn_skill * 3 < 1 / gpGlobals->frametime ? turn_skill * 3 : 1 / gpGlobals->frametime);

      // offset yaw and pitch at some percentage of the deviation, based on skill and frame time.
      pEdict->v.yaw_speed = v_deviation.y * gpGlobals->frametime * speed;
      pEdict->v.pitch_speed = v_deviation.x * gpGlobals->frametime * speed;
   }

   // else do we want new RACC style aiming ?
   else if (turn_method == AIM_RACC)
   {
      // if bot is aiming at something, aim fast, else take our time...
      if (AimsAtSomething (pEdict))
         speed = 0.7 + (turn_skill - 1) / 10; // fast aim
      else
         speed = 0.2 + (turn_skill - 1) / 20; // slow aim

      // thanks Tobias "Killaruna" Heimann and Johannes "@$3.1415rin" Lampel for this one
      pEdict->v.yaw_speed = (pEdict->v.yaw_speed * exp (log (speed / 2) * gpGlobals->frametime * 20)
                             + speed * v_deviation.y * (1 - exp (log (speed / 2) * gpGlobals->frametime * 20)))
                            * gpGlobals->frametime * 20;
      pEdict->v.pitch_speed = (pEdict->v.pitch_speed * exp (log (speed / 2) * gpGlobals->frametime * 20)
                               + speed * v_deviation.x * (1 - exp (log (speed / 2) * gpGlobals->frametime * 20)))
                              * gpGlobals->frametime * 20;

      // influence of y movement on x axis, based on skill (less influence than x on y since it's
      // easier and more natural for the bot to "move its mouse" horizontally than vertically)
      if (pEdict->v.pitch_speed > 0)
         pEdict->v.pitch_speed += pEdict->v.yaw_speed / (1.5 * (1 + turn_skill));
      else
         pEdict->v.pitch_speed -= pEdict->v.yaw_speed / (1.5 * (1 + turn_skill));

      // influence of x movement on y axis, based on skill
      if (pEdict->v.yaw_speed > 0)
         pEdict->v.yaw_speed += pEdict->v.pitch_speed / (1 + turn_skill);
      else
         pEdict->v.yaw_speed -= pEdict->v.pitch_speed / (1 + turn_skill);
   }

   // move the aim cursor
   pEdict->v.v_angle = UTIL_WrapAngles (bot_v_angle[ENTINDEX (pEdict) - 1] + Vector (pEdict->v.pitch_speed, pEdict->v.yaw_speed, 0)); 

   // set the body angles to point the gun correctly
   pEdict->v.angles.x = UTIL_WrapAngle (pEdict->v.v_angle.x / 3);
   pEdict->v.angles.y = UTIL_WrapAngle (pEdict->v.v_angle.y);
   pEdict->v.angles.z = 0;

   // did the bot maker not get rid of Paulo-La-Frite's "bugfix" bug ?
   if (aim_fix)
      pEdict->v.angles.x = UTIL_WrapAngle (-pEdict->v.angles.x); // if the bug is still here, fix it again

   // if we want to save the bot's view angles, remember these view angles
   if (save_angles)
      bot_v_angle[ENTINDEX (pEdict) - 1] = UTIL_WrapAngles (pEdict->v.v_angle);

   return;
}


bool AimsAtSomething (edict_t *pEdict)
{
   TraceResult tr;

   if (FNullEnt (pEdict))
      return FALSE; // reliability check

   // first make vectors in the direction pEdict wants to look
   MAKE_VECTORS (UTIL_WrapAngles (Vector (pEdict->v.idealpitch, pEdict->v.ideal_yaw, 0)));
   TRACE_LINE (pEdict->v.origin + pEdict->v.view_ofs,
               pEdict->v.origin + pEdict->v.view_ofs + gpGlobals->v_forward * 9999,
               dont_ignore_monsters, pEdict, &tr);
   if (strcmp (STRING (tr.pHit->v.classname), "player") == 0)
      return (TRUE); // bot is aiming at a player

   return (FALSE); // bot is aiming at anything but a player
}


float UTIL_AngleOfVectors (Vector v1, Vector v2)
{
   if ((v1.Length () == 0) || (v2.Length () == 0))
      return 0; // avoid zero divide

   return (UTIL_WrapAngle (acos (DotProduct (v1, v2) / (v1.Length () * v2.Length ())) * 180 / M_PI));
}


float UTIL_WrapAngle (float angle_to_wrap)
{
   float angle = angle_to_wrap;

   // check for wraparound of angle
   if (angle >= 180)
      angle -= 360 * ((int) (angle / 360) + 1);
   if (angle < -180)
      angle += 360 * ((int) (angle / 360) + 1);

   return angle;
}


Vector UTIL_WrapAngles (Vector angles_to_wrap)
{
   Vector angles = angles_to_wrap;

   // check for wraparound of angles
   if (angles.x >= 180)
      angles.x -= 360 * ((int) (angles.x / 360) + 1);
   if (angles.x < -180)
      angles.x += 360 * ((int) (angles.x / 360) + 1);
   if (angles.y >= 180)
      angles.y -= 360 * ((int) (angles.y / 360) + 1);
   if (angles.y < -180)
      angles.y += 360 * ((int) (angles.y / 360) + 1);
   if (angles.z >= 180)
      angles.z -= 360 * ((int) (angles.z / 360) + 1);
   if (angles.z < -180)
      angles.z += 360 * ((int) (angles.z / 360) + 1);

   return angles;
}
